@@ -1,7 +1,7 @@
 #!/usr/bin/env python
 
 # pyinotify.py - python interface to inotify
-# Copyright (c) 2010 Sebastien Martini <seb@dbzteam.org>
+# Copyright (c) 2005-2011 Sebastien Martini <seb@dbzteam.org>
 #
 # Permission is hereby granted, free of charge, to any person obtaining a copy
 # of this software and associated documentation files (the "Software"), to deal
@@ -42,28 +42,13 @@ class UnsupportedPythonVersionError(PyinotifyError):
         @param version: Current Python version
         @type version: string
         """
-        PyinotifyError.__init__(self,
-                                ('Python %s is unsupported, requires '
-                                 'at least Python 2.4') % version)
-
-
-class UnsupportedLibcVersionError(PyinotifyError):
-    """
-    Raised on unsupported libc versions.
-    """
-    def __init__(self, version):
-        """
-        @param version: Current Libc version
-        @type version: string
-        """
-        PyinotifyError.__init__(self,
-                                ('Libc %s is not supported, requires '
-                                 'at least Libc 2.4') % version)
+        err = 'Python %s is unsupported, requires at least Python 2.4'
+        PyinotifyError.__init__(self, err % version)
 
 
 # Check Python version
 import sys
-if sys.version < '2.4':
+if sys.version_info < (2, 4):
     raise UnsupportedPythonVersionError(sys.version)
 
 
@@ -81,10 +66,7 @@ import atexit
 from collections import deque
 from datetime import datetime, timedelta
 import time
-import fnmatch
 import re
-import ctypes
-import ctypes.util
 import asyncore
 import glob
 
@@ -93,9 +75,21 @@ try:
 except ImportError:
     pass  # Will fail on Python 2.4 which has reduce() builtin anyway.
 
+try:
+    import ctypes
+    import ctypes.util
+except ImportError:
+    ctypes = None
+
+try:
+    import inotify_syscalls
+except ImportError:
+    inotify_syscalls = None
+
+
 __author__ = "seb@dbzteam.org (Sebastien Martini)"
 
-__version__ = "0.9.0"
+__version__ = "0.9.2"
 
 __metaclass__ = type  # Use new-style classes by default
 
@@ -106,37 +100,157 @@ __metaclass__ = type  # Use new-style classes by default
 COMPATIBILITY_MODE = False
 
 
-# Load libc
-LIBC = None
-STRERRNO = None
-if sys.version_info[0] >= 2 and sys.version_info[1] >= 6:
-    LIBC = ctypes.CDLL(ctypes.util.find_library('c'), use_errno=True)
-    def _strerrno():
-        code = ctypes.get_errno()
-        return ' Errno=%s (%s)' % (os.strerror(code), errno.errorcode[code])
-    STRERRNO = _strerrno
-else:
-    LIBC = ctypes.CDLL(ctypes.util.find_library('c'))
-    STRERRNO = lambda : ''
-
-# The libc version > 2.4 check.
-# XXX: Maybe it is better to check if the libc has the needed functions inside?
-#      Because there are inotify patches for libc 2.3.6.
-LIBC.gnu_get_libc_version.restype = ctypes.c_char_p
-LIBC_VERSION = LIBC.gnu_get_libc_version()
-if (int(LIBC_VERSION.split('.')[0]) < 2 or
-    (int(LIBC_VERSION.split('.')[0]) == 2 and
-     int(LIBC_VERSION.split('.')[1]) < 4)):
-    raise UnsupportedLibcVersionError(LIBC_VERSION)
+class InotifyBindingNotFoundError(PyinotifyError):
+    """
+    Raised when no inotify support couldn't be found.
+    """
+    def __init__(self):
+        err = "Couldn't find any inotify binding"
+        PyinotifyError.__init__(self, err)
 
 
-class PyinotifyLogger(logging.Logger):
+class INotifyWrapper:
+    """
+    Abstract class wrapping access to inotify's functions. This is an
+    internal class.
+    """
+    @staticmethod
+    def create():
+        # First, try to use ctypes.
+        if ctypes:
+            inotify = _CtypesLibcINotifyWrapper()
+            if inotify.init():
+                return inotify
+        # Second, see if C extension is compiled.
+        if inotify_syscalls:
+            inotify = _INotifySyscallsWrapper()
+            if inotify.init():
+                return inotify
+
+    def get_errno(self):
+        """
+        Return None is no errno code is available.
+        """
+        return self._get_errno()
+
+    def str_errno(self):
+        code = self.get_errno()
+        if code is None:
+            return 'Errno: no errno support'
+        return 'Errno=%s (%s)' % (os.strerror(code), errno.errorcode[code])
+
+    def inotify_init(self):
+        return self._inotify_init()
+
+    def inotify_add_watch(self, fd, pathname, mask):
+        # Unicode strings must be encoded to string prior to calling this
+        # method.
+        assert isinstance(pathname, str)
+        return self._inotify_add_watch(fd, pathname, mask)
+
+    def inotify_rm_watch(self, fd, wd):
+        return self._inotify_rm_watch(fd, wd)
+
+
+class _INotifySyscallsWrapper(INotifyWrapper):
+    def __init__(self):
+        # Stores the last errno value.
+        self._last_errno = None
+
+    def init(self):
+        assert inotify_syscalls
+        return True
+
+    def _get_errno(self):
+        return self._last_errno
+
+    def _inotify_init(self):
+        try:
+            fd = inotify_syscalls.inotify_init()
+        except IOError, err:
+            self._last_errno = err.errno
+            return -1
+        return fd
+
+    def _inotify_add_watch(self, fd, pathname, mask):
+        try:
+            wd = inotify_syscalls.inotify_add_watch(fd, pathname, mask)
+        except IOError, err:
+            self._last_errno = err.errno
+            return -1
+        return wd
+
+    def _inotify_rm_watch(self, fd, wd):
+        try:
+            ret = inotify_syscalls.inotify_rm_watch(fd, wd)
+        except IOError, err:
+            self._last_errno = err.errno
+            return -1
+        return ret
+
+
+class _CtypesLibcINotifyWrapper(INotifyWrapper):
+    def __init__(self):
+        self._libc = None
+        self._get_errno_func = None
+
+    def init(self):
+        assert ctypes
+        libc_name = None
+        try:
+            libc_name = ctypes.util.find_library('c')
+        except (OSError, IOError):
+            pass  # Will attemp to load it with None anyway.
+
+        if sys.version_info >= (2, 6):
+            self._libc = ctypes.CDLL(libc_name, use_errno=True)
+            self._get_errno_func = ctypes.get_errno
+        else:
+            self._libc = ctypes.CDLL(libc_name)
+            try:
+                location = self._libc.__errno_location
+                location.restype = ctypes.POINTER(ctypes.c_int)
+                self._get_errno_func = lambda: location().contents.value
+            except AttributeError:
+                pass
+
+        # Eventually check that libc has needed inotify bindings.
+        if (not hasattr(self._libc, 'inotify_init') or
+            not hasattr(self._libc, 'inotify_add_watch') or
+            not hasattr(self._libc, 'inotify_rm_watch')):
+            return False
+        return True
+
+    def _get_errno(self):
+        if self._get_errno_func is not None:
+            return self._get_errno_func()
+        return None
+
+    def _inotify_init(self):
+        assert self._libc is not None
+        return self._libc.inotify_init()
+
+    def _inotify_add_watch(self, fd, pathname, mask):
+        assert self._libc is not None
+        pathname = ctypes.create_string_buffer(pathname)
+        return self._libc.inotify_add_watch(fd, pathname, mask)
+
+    def _inotify_rm_watch(self, fd, wd):
+        assert self._libc is not None
+        return self._libc.inotify_rm_watch(fd, wd)
+
+    def _sysctl(self, *args):
+        assert self._libc is not None
+        return self._libc.sysctl(*args)
+
+
+class _PyinotifyLogger(logging.Logger):
     """
     Pyinotify logger used for logging unicode strings.
     """
     def makeRecord(self, name, level, fn, lno, msg, args, exc_info, func=None,
                    extra=None):
-        rv = UnicodeLogRecord(name, level, fn, lno, msg, args, exc_info, func)
+        rv = _UnicodeLogRecord(name, level, fn, lno, msg, args, exc_info, func)
         if extra is not None:
             for key in extra:
                 if (key in ["message", "asctime"]) or (key in rv.__dict__):
@@ -145,7 +259,7 @@ class PyinotifyLogger(logging.Logger):
         return rv
 
 
-class UnicodeLogRecord(logging.LogRecord):
+class _UnicodeLogRecord(logging.LogRecord):
     def __init__(self, name, level, pathname, lineno,
                  msg, args, exc_info, func=None):
         py_version = sys.version_info
@@ -183,11 +297,11 @@ class UnicodeLogRecord(logging.LogRecord):
 # Logging
 def logger_init():
     """Initialize logger instance."""
-    logging.setLoggerClass(PyinotifyLogger)
+    logging.setLoggerClass(_PyinotifyLogger)
     log = logging.getLogger("pyinotify")
     console_handler = logging.StreamHandler()
     console_handler.setFormatter(
-        logging.Formatter("[Pyinotify %(levelname)s] %(message)s"))
+        logging.Formatter("[%(asctime)s %(name)s %(levelname)s] %(message)s"))
     log.addHandler(console_handler)
     log.setLevel(20)
     return log
@@ -210,11 +324,27 @@ class SysCtlINotify:
                      'max_user_watches': 2,
                      'max_queued_events': 3}
 
-    def __init__(self, attrname):
-        sino = ctypes.c_int * 3
+    def __init__(self, attrname, inotify_wrapper):
+        # FIXME: right now only supporting ctypes
+        assert ctypes
         self._attrname = attrname
+        self._inotify_wrapper = inotify_wrapper
+        sino = ctypes.c_int * 3
         self._attr = sino(5, 20, SysCtlINotify.inotify_attrs[attrname])
 
+    @staticmethod
+    def create(attrname):
+        """
+        Factory method instanciating and returning the right wrapper.
+        """
+        # FIXME: right now only supporting ctypes
+        if ctypes is None:
+            return None
+        inotify_wrapper = _CtypesLibcINotifyWrapper()
+        if not inotify_wrapper.init():
+            return None
+        return SysCtlINotify(attrname, inotify_wrapper)
+
     def get_val(self):
         """
         Gets attribute's value.
@@ -224,10 +354,10 @@ class SysCtlINotify:
         """
         oldv = ctypes.c_int(0)
         size = ctypes.c_int(ctypes.sizeof(oldv))
-        LIBC.sysctl(self._attr, 3,
-                    ctypes.c_voidp(ctypes.addressof(oldv)),
-                    ctypes.addressof(size),
-                    None, 0)
+        self._inotify_wrapper._sysctl(self._attr, 3,
+                                      ctypes.c_voidp(ctypes.addressof(oldv)),
+                                      ctypes.addressof(size),
+                                      None, 0)
         return oldv.value
 
     def set_val(self, nval):
@@ -241,11 +371,11 @@ class SysCtlINotify:
         sizeo = ctypes.c_int(ctypes.sizeof(oldv))
         newv = ctypes.c_int(nval)
         sizen = ctypes.c_int(ctypes.sizeof(newv))
-        LIBC.sysctl(self._attr, 3,
-                    ctypes.c_voidp(ctypes.addressof(oldv)),
-                    ctypes.addressof(sizeo),
-                    ctypes.c_voidp(ctypes.addressof(newv)),
-                    ctypes.addressof(sizen))
+        self._inotify_wrapper._sysctl(self._attr, 3,
+                                      ctypes.c_voidp(ctypes.addressof(oldv)),
+                                      ctypes.addressof(sizeo),
+                                      ctypes.c_voidp(ctypes.addressof(newv)),
+                                      ctypes.addressof(sizen))
 
     value = property(get_val, set_val)
 
@@ -253,13 +383,16 @@ class SysCtlINotify:
         return '<%s=%d>' % (self._attrname, self.get_val())
 
 
-# Singleton instances
+# Inotify's variables
+#
+# FIXME: currently these variables are only accessible when ctypes is used,
+#        otherwise there are set to None.
 #
 # read: myvar = max_queued_events.value
 # update: max_queued_events.value = 42
 #
 for attrname in ('max_queued_events', 'max_user_instances', 'max_user_watches'):
-    globals()[attrname] = SysCtlINotify(attrname)
+    globals()[attrname] = SysCtlINotify.create(attrname)
 
 
 class EventsCodes:
@@ -424,16 +557,19 @@ class _Event:
                 value = hex(getattr(self, attr))
             elif isinstance(value, basestring) and not value:
                 value = "''"
-            s += ' %s%s%s' % (Color.field_name(attr),
-                              Color.punctuation('='),
-                              Color.field_value(value))
+            s += ' %s%s%s' % (output_format.field_name(attr),
+                              output_format.punctuation('='),
+                              output_format.field_value(value))
 
-        s = '%s%s%s %s' % (Color.punctuation('<'),
-                           Color.class_name(self.__class__.__name__),
+        s = '%s%s%s %s' % (output_format.punctuation('<'),
+                           output_format.class_name(self.__class__.__name__),
                            s,
-                           Color.punctuation('>'))
+                           output_format.punctuation('>'))
         return s
 
+    def __str__(self):
+        return repr(self)
+
 
 class _RawEvent(_Event):
     """
@@ -454,19 +590,20 @@ class _RawEvent(_Event):
                      on the watched item itself.
         @type name: string or None
         """
-        # name: remove trailing '\0'
-        _Event.__init__(self, {'wd': wd,
-                               'mask': mask,
-                               'cookie': cookie,
-                               'name': name.rstrip('\0')})
-        log.debug(repr(self))
-        # Use this variable to cache the result of str(self)
+        # Use this variable to cache the result of str(self), this object
+        # is immutable.
         self._str = None
+        # name: remove trailing '\0'
+        d = {'wd': wd,
+             'mask': mask,
+             'cookie': cookie,
+             'name': name.rstrip('\0')}
+        _Event.__init__(self, d)
+        log.debug(str(self))
 
     def __str__(self):
         if self._str is None:
-            self._str = '%s %s %s %s' % (str(self.wd), str(self.mask),
-                                         str(self.cookie), self.name)
+            self._str = _Event.__str__(self)
         return self._str
 
 
@@ -625,22 +762,28 @@ class _SysProcessEvent(_ProcessEvent):
                                 rec=False, auto_add=watch_.auto_add,
                                 exclude_filter=watch_.exclude_filter)
 
-                # Trick to handle mkdir -p /t1/t2/t3 where t1 is watched and
-                # t2 and t3 are created.
-                # Since the directory is new, then everything inside it
-                # must also be new.
+                # Trick to handle mkdir -p /d1/d2/t3 where d1 is watched and
+                # d2 and t3 (directory or file) are created.
+                # Since the directory d2 is new, then everything inside it must
+                # also be new.
                 created_dir_wd = addw_ret.get(created_dir)
-                if (created_dir_wd is not None) and created_dir_wd > 0:
+                if (created_dir_wd is not None) and (created_dir_wd > 0):
                     for name in os.listdir(created_dir):
                         inner = os.path.join(created_dir, name)
-                        if (os.path.isdir(inner) and
-                            self._watch_manager.get_wd(inner) is None):
-                            # Generate (simulate) creation event for sub
-                            # directories.
-                            rawevent = _RawEvent(created_dir_wd,
-                                                 IN_CREATE | IN_ISDIR,
-                                                 0, name)
-                            self._notifier.append_event(rawevent)
+                        if self._watch_manager.get_wd(inner) is not None:
+                            continue
+                        # Generate (simulate) creation events for sub-
+                        # directories and files.
+                        if os.path.isfile(inner):
+                            # symlinks are handled as files.
+                            flags = IN_CREATE
+                        elif os.path.isdir(inner):
+                            flags = IN_CREATE | IN_ISDIR
+                        else:
+                            # This path should not be taken.
+                            continue
+                        rawevent = _RawEvent(created_dir_wd, flags, 0, name)
+                        self._notifier.append_event(rawevent)
         return self.process_default(raw_event)
 
     def process_IN_MOVED_FROM(self, raw_event):
@@ -884,8 +1027,9 @@ class PrintAllEvents(ProcessEvent):
                       IN_Q_OVERFLOW events (see method process_IN_Q_OVERFLOW).
         @type event: Event instance
         """
-        self._out.write(repr(event))
+        self._out.write(str(event))
         self._out.write('\n')
+        self._out.flush()
 
 
 class ChainIfTrue(ProcessEvent):
@@ -952,24 +1096,24 @@ class Stats(ProcessEvent):
 
         l = []
         for ev, value in sorted(stats.items(), key=lambda x: x[0]):
-            l.append(' %s=%s' % (Color.field_name(ev),
-                                 Color.field_value(value)))
-        s = '<%s%s >' % (Color.class_name(self.__class__.__name__),
+            l.append(' %s=%s' % (output_format.field_name(ev),
+                                 output_format.field_value(value)))
+        s = '<%s%s >' % (output_format.class_name(self.__class__.__name__),
                          ''.join(l))
         return s
 
     def dump(self, filename):
         """
-        Dumps statistics to file |filename|.
+        Dumps statistics.
 
-        @param filename: pathname.
+        @param filename: filename where stats will be dumped, filename is
+                         created and must not exist prior to this call.
         @type filename: string
         """
-        file_obj = file(filename, 'wb')
-        try:
-            file_obj.write(str(self))
-        finally:
-            file_obj.close()
+        flags = os.O_WRONLY|os.O_CREAT|os.O_NOFOLLOW|os.O_EXCL
+        fd = os.open(filename, flags, 0600)
+        os.write(fd, str(self))
+        os.close(fd)
 
     def __str__(self, scale=45):
         stats = self._stats_copy()
@@ -978,12 +1122,12 @@ class Stats(ProcessEvent):
 
         m = max(stats.values())
         unity = float(scale) / m
-        fmt = '%%-26s%%-%ds%%s' % (len(Color.field_value('@' * scale))
+        fmt = '%%-26s%%-%ds%%s' % (len(output_format.field_value('@' * scale))
                                    + 1)
         def func(x):
-            return fmt % (Color.field_name(x[0]),
-                          Color.field_value('@' * int(x[1] * unity)),
-                          Color.simple('%d' % x[1], 'yellow'))
+            return fmt % (output_format.field_name(x[0]),
+                          output_format.field_value('@' * int(x[1] * unity)),
+                          output_format.simple('%d' % x[1], 'yellow'))
         s = '\n'.join(map(func, sorted(stats.items(), key=lambda x: x[0])))
         return s
 
@@ -1189,42 +1333,25 @@ class Notifier:
         if self._coalesce:
             self._eventset.clear()
 
-    def __daemonize(self, pid_file=None, force_kill=False, stdin=os.devnull,
-                    stdout=os.devnull, stderr=os.devnull):
+    def __daemonize(self, pid_file=None, stdin=os.devnull, stdout=os.devnull,
+                    stderr=os.devnull):
         """
-        pid_file: file to which the pid will be written.
-        force_kill: if True kill the process associated to pid_file.
-        stdin, stdout, stderr: files associated to common streams.
+        @param pid_file: file where the pid will be written. If pid_file=None
+                         the pid is written to
+                         /var/run/<sys.argv[0]|pyinotify>.pid, if pid_file=False
+                         no pid_file is written.
+        @param stdin:
+        @param stdout:
+        @param stderr: files associated to common streams.
         """
         if pid_file is None:
             dirname = '/var/run/'
             basename = os.path.basename(sys.argv[0]) or 'pyinotify'
             pid_file = os.path.join(dirname, basename + '.pid')
 
-        if os.path.exists(pid_file):
-            fo = file(pid_file, 'rb')
-            try:
-                try:
-                    pid = int(fo.read())
-                except ValueError:
-                    pid = None
-                if pid is not None:
-                    try:
-                        os.kill(pid, 0)
-                    except OSError, err:
-                        if err.errno == errno.ESRCH:
-                            log.debug(err)
-                        else:
-                            log.error(err)
-                    else:
-                        if not force_kill:
-                            s = 'There is already a pid file %s with pid %d'
-                            raise NotifierError(s % (pid_file, pid))
-                        else:
-                            os.kill(pid, 9)
-            finally:
-                fo.close()
-
+        if pid_file != False and os.path.lexists(pid_file):
+            err = 'Cannot daemonize: pid file %s already exists.' % pid_file
+            raise NotifierError(err)
 
         def fork_daemon():
             # Adapted from Chad J. Schroeder's recipe
@@ -1237,7 +1364,7 @@ class Notifier:
                 if (pid == 0):
                     # child
                     os.chdir('/')
-                    os.umask(0)
+                    os.umask(022)
                 else:
                     # parent 2
                     os._exit(0)
@@ -1247,22 +1374,22 @@ class Notifier:
 
             fd_inp = os.open(stdin, os.O_RDONLY)
             os.dup2(fd_inp, 0)
-            fd_out = os.open(stdout, os.O_WRONLY|os.O_CREAT)
+            fd_out = os.open(stdout, os.O_WRONLY|os.O_CREAT, 0600)
             os.dup2(fd_out, 1)
-            fd_err = os.open(stderr, os.O_WRONLY|os.O_CREAT)
+            fd_err = os.open(stderr, os.O_WRONLY|os.O_CREAT, 0600)
             os.dup2(fd_err, 2)
 
         # Detach task
         fork_daemon()
 
         # Write pid
-        file_obj = file(pid_file, 'wb')
-        try:
-            file_obj.write(str(os.getpid()) + '\n')
-        finally:
-            file_obj.close()
-
-        atexit.register(lambda : os.unlink(pid_file))
+        if pid_file != False:
+            flags = os.O_WRONLY|os.O_CREAT|os.O_NOFOLLOW|os.O_EXCL
+            fd_pid = os.open(pid_file, flags, 0600)
+            os.write(fd_pid, str(os.getpid()) + '\n')
+            os.close(fd_pid)
+            # Register unlink function
+            atexit.register(lambda : os.unlink(pid_file))
 
 
     def _sleep(self, ref_time):
@@ -1292,7 +1419,12 @@ class Notifier:
         @type daemonize: boolean
         @param args: Optional and relevant only if daemonize is True. Remaining
                      keyworded arguments are directly passed to daemonize see
-                     __daemonize() method.
+                     __daemonize() method. If pid_file=None or is set to a
+                     pathname the caller must ensure the file does not exist
+                     before this method is called otherwise an exception
+                     pyinotify.NotifierError will be raised. If pid_file=False
+                     it is still daemonized but the pid is not written in any
+                     file.
         @type args: various
         """
         if daemonize:
@@ -1449,6 +1581,9 @@ class Watch:
     Represent a watch, i.e. a file or directory being watched.
 
     """
+    __slots__ = ('wd', 'path', 'mask', 'proc_fun', 'auto_add',
+                 'exclude_filter', 'dir')
+
     def __init__(self, wd, path, mask, proc_fun, auto_add, exclude_filter):
         """
         Initializations.
@@ -1481,15 +1616,16 @@ class Watch:
         @return: String representation.
         @rtype: str
         """
-        s = ' '.join(['%s%s%s' % (Color.field_name(attr),
-                                  Color.punctuation('='),
-                                  Color.field_value(getattr(self, attr))) \
-                      for attr in self.__dict__ if not attr.startswith('_')])
+        s = ' '.join(['%s%s%s' % (output_format.field_name(attr),
+                                  output_format.punctuation('='),
+                                  output_format.field_value(getattr(self,
+                                                                    attr))) \
+                      for attr in self.__slots__ if not attr.startswith('_')])
 
-        s = '%s%s %s %s' % (Color.punctuation('<'),
-                            Color.class_name(self.__class__.__name__),
+        s = '%s%s %s %s' % (output_format.punctuation('<'),
+                            output_format.class_name(self.__class__.__name__),
                             s,
-                            Color.punctuation('>'))
+                            output_format.punctuation('>'))
         return s
 
 
@@ -1581,7 +1717,9 @@ class WatchManager:
     def __init__(self, exclude_filter=lambda path: False):
         """
         Initialization: init inotify, init watch manager dictionary.
-        Raise OSError if initialization fails.
+        Raise OSError if initialization fails, raise InotifyBindingNotFoundError
+        if no inotify binding was found (through ctypes or from direct access to
+        syscalls).
 
         @param exclude_filter: boolean function, returns True if current
                                path must be excluded from being watched.
@@ -1591,10 +1729,27 @@ class WatchManager:
         """
         self._exclude_filter = exclude_filter
         self._wmd = {}  # watch dict key: watch descriptor, value: watch
-        self._fd = LIBC.inotify_init() # inotify's init, file descriptor
+
+        self._inotify_wrapper = INotifyWrapper.create()
+        if self._inotify_wrapper is None:
+            raise InotifyBindingNotFoundError()
+
+        self._fd = self._inotify_wrapper.inotify_init() # file descriptor
         if self._fd < 0:
-            err = 'Cannot initialize new instance of inotify%s' % STRERRNO()
-            raise OSError(err)
+            err = 'Cannot initialize new instance of inotify, %s'
+            raise OSError(err % self._inotify_wrapper.str_errno())
+
+    def close(self):
+        """
+        Close inotify's file descriptor, this action will also automatically
+        remove (i.e. stop watching) all its associated watch descriptors.
+        After a call to this method the WatchManager's instance become useless
+        and cannot be reused, a new instance must then be instanciated. It
+        makes sense to call this method in few situations for instance if
+        several independant WatchManager must be instanciated or if all watches
+        must be removed and no other watches need to be added.
+        """
+        os.close(self._fd)
 
     def get_fd(self):
         """
@@ -1640,8 +1795,8 @@ class WatchManager:
         """
         Format path to its internal (stored in watch manager) representation.
         """
-        # Unicode strings are converted to byte strings, it seems to be
-        # required because LIBC.inotify_add_watch does not work well when
+        # Unicode strings are converted back to strings, because it seems
+        # that inotify_add_watch from ctypes does not work well when
         # it receives an ctypes.create_unicode_buffer instance as argument.
         # Therefore even wd are indexed with bytes string and not with
         # unicode paths.
@@ -1654,17 +1809,15 @@ class WatchManager:
         Add a watch on path, build a Watch object and insert it in the
         watch manager dictionary. Return the wd value.
         """
-        byte_path = self.__format_path(path)
-        wd_ = LIBC.inotify_add_watch(self._fd,
-                                     ctypes.create_string_buffer(byte_path),
-                                     mask)
-        if wd_ < 0:
-            return wd_
-        watch_ = Watch(wd=wd_, path=byte_path, mask=mask, proc_fun=proc_fun,
-                       auto_add=auto_add, exclude_filter=exclude_filter)
-        self._wmd[wd_] = watch_
-        log.debug('New %s', watch_)
-        return wd_
+        path = self.__format_path(path)
+        wd = self._inotify_wrapper.inotify_add_watch(self._fd, path, mask)
+        if wd < 0:
+            return wd
+        watch = Watch(wd=wd, path=path, mask=mask, proc_fun=proc_fun,
+                      auto_add=auto_add, exclude_filter=exclude_filter)
+        self._wmd[wd] = watch
+        log.debug('New %s', watch)
+        return wd
 
     def __glob(self, path, do_glob):
         if do_glob:
@@ -1743,8 +1896,9 @@ class WatchManager:
                                                             auto_add,
                                                             exclude_filter)
                         if wd < 0:
-                            err = 'add_watch: cannot watch %s WD=%d%s'
-                            err = err % (rpath, wd, STRERRNO())
+                            err = ('add_watch: cannot watch %s WD=%d, %s' % \
+                                       (rpath, wd,
+                                        self._inotify_wrapper.str_errno()))
                             if quiet:
                                 log.error(err)
                             else:
@@ -1835,12 +1989,12 @@ class WatchManager:
                 raise WatchManagerError(err, ret_)
 
             if mask:
-                addw = LIBC.inotify_add_watch
-                wd_ = addw(self._fd, ctypes.create_string_buffer(apath), mask)
+                wd_ = self._inotify_wrapper.inotify_add_watch(self._fd, apath,
+                                                              mask)
                 if wd_ < 0:
                     ret_[awd] = False
-                    err = 'update_watch: cannot update %s WD=%d%s'
-                    err = err % (apath, wd_, STRERRNO())
+                    err = ('update_watch: cannot update %s WD=%d, %s' % \
+                               (apath, wd_, self._inotify_wrapper.str_errno()))
                     if quiet:
                         log.error(err)
                         continue
@@ -1946,15 +2100,19 @@ class WatchManager:
         ret_ = {}  # return {wd: bool, ...}
         for awd in lwd:
             # remove watch
-            wd_ = LIBC.inotify_rm_watch(self._fd, awd)
+            wd_ = self._inotify_wrapper.inotify_rm_watch(self._fd, awd)
             if wd_ < 0:
                 ret_[awd] = False
-                err = 'rm_watch: cannot remove WD=%d%s' % (awd, STRERRNO())
+                err = ('rm_watch: cannot remove WD=%d, %s' % \
+                           (awd, self._inotify_wrapper.str_errno()))
                 if quiet:
                     log.error(err)
                     continue
                 raise WatchManagerError(err, ret_)
 
+            # Remove watch from our dictionary
+            if awd in self._wmd:
+                del self._wmd[awd]
             ret_[awd] = True
             log.debug('Watch WD=%d (%s) removed', awd, self.get_path(awd))
         return ret_
@@ -2002,54 +2160,55 @@ class WatchManager:
                               exclude_filter=lambda path: False)
 
 
-class Color:
+class RawOutputFormat:
     """
-    Internal class. Provide fancy colors used by string representations.
+    Format string representations.
     """
-    normal = "\033[0m"
-    black = "\033[30m"
-    red = "\033[31m"
-    green = "\033[32m"
-    yellow = "\033[33m"
-    blue = "\033[34m"
-    purple = "\033[35m"
-    cyan = "\033[36m"
-    bold = "\033[1m"
-    uline = "\033[4m"
-    blink = "\033[5m"
-    invert = "\033[7m"
+    def __init__(self, format=None):
+        self.format = format or {}
 
-    @staticmethod
-    def punctuation(s):
+    def simple(self, s, attribute):
+        if not isinstance(s, str):
+            s = str(s)
+        return (self.format.get(attribute, '') + s +
+                self.format.get('normal', ''))
+
+    def punctuation(self, s):
         """Punctuation color."""
-        return Color.normal + s + Color.normal
+        return self.simple(s, 'normal')
 
-    @staticmethod
-    def field_value(s):
+    def field_value(self, s):
         """Field value color."""
-        if not isinstance(s, basestring):
-            s = str(s)
-        return Color.purple + s + Color.normal
+        return self.simple(s, 'purple')
 
-    @staticmethod
-    def field_name(s):
+    def field_name(self, s):
         """Field name color."""
-        return Color.blue + s + Color.normal
+        return self.simple(s, 'blue')
 
-    @staticmethod
-    def class_name(s):
+    def class_name(self, s):
         """Class name color."""
-        return Color.red + Color.bold + s + Color.normal
+        return self.format.get('red', '') + self.simple(s, 'bold')
 
-    @staticmethod
-    def simple(s, color):
-        if not isinstance(s, basestring):
-            s = str(s)
-        try:
-            color_attr = getattr(Color, color)
-        except AttributeError:
-            return s
-        return color_attr + s + Color.normal
+output_format = RawOutputFormat()
+
+class ColoredOutputFormat(RawOutputFormat):
+    """
+    Format colored string representations.
+    """
+    def __init__(self):
+        f = {'normal': '\033[0m',
+             'black': '\033[30m',
+             'red': '\033[31m',
+             'green': '\033[32m',
+             'yellow': '\033[33m',
+             'blue': '\033[34m',
+             'purple': '\033[35m',
+             'cyan': '\033[36m',
+             'bold': '\033[1m',
+             'uline': '\033[4m',
+             'blink': '\033[5m',
+             'invert': '\033[7m'}
+        RawOutputFormat.__init__(self, f)
 
 
 def compatibility_mode():
@@ -2095,12 +2254,24 @@ def command_line():
     parser.add_option("-s", "--stats", action="store_true",
                       dest="stats",
                       help="Display dummy statistics")
+    parser.add_option("-V", "--version", action="store_true",
+                      dest="version",  help="Pyinotify version")
+    parser.add_option("-f", "--raw-format", action="store_true",
+                      dest="raw_format",
+                      help="Disable enhanced output format.")
 
     (options, args) = parser.parse_args()
 
     if options.verbose:
         log.setLevel(10)
 
+    if options.version:
+        print(__version__)
+
+    if not options.raw_format:
+        global output_format
+        output_format = ColoredOutputFormat()
+
     if len(args) < 1:
         path = '/tmp'  # default watched path
     else:
@@ -2132,8 +2303,11 @@ def command_line():
     cb_fun = None
     if options.stats:
         def cb(s):
-            print('%s\n%s\n' % (repr(s.proc_fun()),
-                                s.proc_fun()))
+            sys.stdout.write(repr(s.proc_fun()))
+            sys.stdout.write('\n')
+            sys.stdout.write(str(s.proc_fun()))
+            sys.stdout.write('\n')
+            sys.stdout.flush()
         cb_fun = cb
 
     log.debug('Start monitoring %s, (press c^c to halt pyinotify)' % path)
